/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2003-2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

Components.utils.import("chrome://dict/content/dictclient.js");
Components.utils.import("chrome://dict/content/utf8.js");

//
// This file contains a protocol handler for supplying a channel to the browser
// when an dict:// URL is used.
//

// DictProtocol namespace
if ("undefined" == typeof(DictProtocol)) {
    var DictProtocol = {};
};

DictProtocol.__defineGetter__("CLASS_ID", function() {
    return Components.ID('{d8b60e94-9847-40d4-b8dc-15ee9706029a}');
});


// classes/services defined in this file.
const DICTPROT_HANDLER_CONTRACTID = "@mozilla.org/network/protocol;1?name=dict";

// components used in this file 
const STANDARDURL_CONTRACTID = "@mozilla.org/network/standard-url;1";
const IOSERVICE_CONTRACTID   = "@mozilla.org/network/io-service;1";
const STRINGINPUT_CONTRACTID = "@mozilla.org/io/string-input-stream;1";

// interfaces used in this file
const nsIProtocolHandler   = Components.interfaces.nsIProtocolHandler;
const nsIURI               = Components.interfaces.nsIURI;
const nsIChannel           = Components.interfaces.nsIChannel;
const nsIRequest           = Components.interfaces.nsIRequest;
const nsIIOService         = Components.interfaces.nsIIOService;
const nsISupports          = Components.interfaces.nsISupports;
const nsIInputStream       = Components.interfaces.nsIInputStream;
const nsIStringInputStream = Components.interfaces.nsIStringInputStream;


// DictProtocol.Listener class, which is passed in as the "listener" parameter
// to DictClient::defineWord().
DictProtocol.Listener = function(channel, responseIndex, host, port, database) {
    this._channel       = channel;
    this._responseIndex = responseIndex;

    this._host     = host;
    this._port     = port;
    this._database = database;
}

DictProtocol.Listener.prototype = 
{
    // wraps HTML inside HTML/BODY tag and adds a title
    createHtmlDocument : function (title, body)
    {
        var htmlDocument;
    
        htmlDocument = 
            '<!DOCTYPE html>\n' +
            '<HTML>\n'    +
            '  <HEAD>\n'  + 
            '      <TITLE>' + title + '</TITLE>\n' +
            '      <META http-equiv="Content-Type" content="text/html; charset=utf-8">\n' +
            '      <SCRIPT type="text/javascript">\n' + 

            // This HTML must be self-resizing.  Doing it from the code that
            // creates the window doesn't work with data URIs.
            'function load() {\n' +
            '    window.sizeToContent();\n' +
            '    if (window.outerWidth < window.innerWidth) {\n' +

                     // The window became narrower than the content
                     // This should never happen, but it may be happening
                     // because of bug #7097--a bug in DownloadStatusBar
                     // when it has over 40 items.
                     // I cannot fix DownloadStatusBar, but I can be defensive
                     // and resize the window back to its original width.
            '        window.outerWidth = window.innerWidth;\n' +
            '    }\n' +

                 // If the page is too long to display without scrollbars, then shorten it
            '    var proposedHeight = window.innerHeight + 70;\n' + 
            '    if (600 < proposedHeight) {\n' +
            '        window.innerHeight = 600;\n' + 

                     // Make the window wider to account for the width of the
                     // vertical scrollbar.  If we don't do this, a 
                     // horizontal scrollbar will appear.
            '        window.innerWidth += 25;\n' +
            '    }\n' +

            '}\n' +
            '\n' +
            'function crossReference(uri) {\n' +
            '    window.open(uri, \'_blank\', \'status=no,toolbar=no,menubar=no,location=no,scrollbars=yes,resizable=yes,width=475,height=600\');\n' +
            '    return false;\n' +
            '}\n' +
            '      </SCRIPT>\n' +
            '  </HEAD>\n' +
            '  <BODY onload="load();">\n'  +
            body +
            '  </BODY>\n' +
            '</HTML>\n';
        

        return htmlDocument;
    },

    createHtmlDefinition : function (definition)
    {
        var html = definition;
        
        // normalize all definitions so that they end in a single newline
        html = html.replace(/\s*$/, '\n');

        // Escape HTML entities
        html = html.replace(/&/g, '&amp;');
        html = html.replace(/</g, '&lt;');

        // Words that show up between backslashes are generally
        // pronunciation guides.  See:
        // http://goodies.xfce.org/projects/applications/xfce4-dict
        html = html.replace(
            /\\(\S*?)\\/g,
            '<SPAN style="color:rgb(0,127,0)"><SPAN class="dict-pronunciaton">' +
                '$1' +
                '</SPAN></SPAN>');

        // Words that are cross-referenced by WordNet show up 
        // in braces. e.g., "{word}" and "{word(P)}".
        // Turn these into links.
        var uriBase = 'dict://' + this._host + ':' + this._port + '/d:';
        html = html.replace(
            /{(.*?)(\(.*\))?}/g,
            function(entireMatch, word, hint) {
                var uri = uriBase + encodeURIComponent(dictUtf8ToUcs4(word));
                return '<a ' +
                    'href="' + uri + '" ' +
                    'onclick="return crossReference(\'' + uri + '\');"' +
                    '>' + word + '</a>';
                });

        return html;
    },


    //
    // The next three functions are the DictListener interface
    //

    // Called when the DictClient has definitions
    onHaveDefinitions : function (word, definitions)
    {
        // Format the definitions into HTML
        var htmlBody = '<DIV class="dict-page">\n';

        // count the number of sources -- used for formatting
        var totalSources = 0;
        for (source in definitions) {
            totalSources++;
        }

        var sourceIndex = 0;
        for (source in definitions) {

            htmlBody += '<B><DIV class="dict-database">';
            htmlBody += source;
            htmlBody += '</DIV></B>\n';
            
            htmlBody += '<CODE style="white-space:pre;"><SPAN class="dict-definition">';

            var i;
            for (i = 0; i < definitions[source].length; i++) {

                if (this._responseIndex) {
                    // the caller is only requesting one definition

                    var thisIndex = i + 1; // DICT definitions are 1-indexed
                    if (this._responseIndex == thisIndex) {
                        htmlBody += this.createHtmlDefinition(definitions[source][i].text);
                    }
                } 
                else {
                    // the caller is requesting all definitions
                    var deftext = this.createHtmlDefinition(definitions[source][i].text);

                    // append an extra newline after everything but the
                    // the final definition from the final source.
                    // (Otherwise there'd be a strange-looking blank space
                    // at the bottom of the page.)
                    if (sourceIndex + 1 != totalSources ||
                        i + 1           != definitions[source].length) {
                        deftext += '\n';
                    }

                    htmlBody += deftext;
                }
            }

            htmlBody += '</SPAN></CODE>\n';
            sourceIndex++;
        }

        htmlBody += '</DIV>\n';

        // create an HTML document
        var htmlDocument = this.createHtmlDocument(word, htmlBody);
        
        // give the HTML document to the channel
        this._channel.Write(htmlDocument);
    },

    createHtmlMatch : function (match)
    {
        var quotelessMatch = match.replace(/(^")|("$)/g, '');

        var uriSpec = 'dict://' + this._host;

        // only set non-default ports
        if (this._port != 2628) {
            uriSpec += ':' + this._port;
        }

        // URI escape the word
        uriSpec += '/d:' + escape(quotelessMatch);

        // only set non-default databases
        if (this._database != '!') {
            uriSpec += ':' + this._database;
        }

        return '<A href="' + uriSpec + '">' + quotelessMatch + '</a><BR>\n';
    },

    // Called when the DictClient has closest matches
    onHaveClosestMatches : function (word, matches)
    {
        // Display all close matches all and let the user decide which to define
        var htmlBody = 
            '<DIV class="dict-page" style="width:450">\n' + 
                '<B><DIV class="dict-heading">Closest Matches</DIV></B><BR>\n' + 
                '<DIV class="dict-text">\n' + 
                    'Some matches for "' + word + '" were found.<BR>\n' +
                    'To define one of them, click on a link below.' +
                 '</DIV>\n';

        // add a spacer between the text and list of words
        htmlBody += '<HR class="dict-horizontal-rule">\n';

        var thisIndex = 1; // DICT matches are 1-indexed
        for (match in matches) {

            if (this._responseIndex) {
                // the caller is only requesting one match
                if (this._responseIndex == thisIndex) {
                    htmlBody += this.createHtmlMatch(match);
                }
            }
            else {
                // the caller is requesting all matches
                htmlBody += this.createHtmlMatch(match);
            }

            thisIndex++; // DICT matches are 1-indexed
        }

        htmlBody += '</DIV>\n';

        var htmlDocument = this.createHtmlDocument(
            word + ' - Closest Matches',
            htmlBody);

        // give the HTML document to the channel
        this._channel.Write(htmlDocument);
    },

    // called when the DictClient could not find any definitions
    onNoDefinitionsFound : function (word) 
    {
        // create an HTML document
        var htmlDocument = this.createHtmlDocument(
            'No Definitions Found', 
            'No definitions were found for "' + word + '".\n');

        // give the HTML document to the channel
        this._channel.Write(htmlDocument);
    },

    // called when the DictClient could not find any matches
    onNoMatchesFound : function (word) 
    {
        // create an HTML document
        var htmlDocument = this.createHtmlDocument(
            'No Matches Found', 
            'No matches were found for "' + word + '".\n');

        // give the HTML document to the channel
        this._channel.Write(htmlDocument);
    },

    // called when the DictClient encounters some unrecoverable error
    onError : function (error)
    {
        // create an HTML document
        var htmlDocument = this.createHtmlDocument(
            'Error', 
            '<PRE>' + error + '</PRE>');
        
        // give the HTML document to the channel
        this._channel.Write(htmlDocument);
    }
}


// Channel used by the DictProtocol
DictProtocol.Channel = function(input_uri)
{
    this.URI         = input_uri;
    this.originalURI = input_uri;
}

DictProtocol.Channel.prototype = 
{
    // nsISupports
    QueryInterface : function (iid)
    {
        if (!iid.equals(nsIChannel) && 
            !iid.equals(nsIRequest) &&
            !iid.equals(nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }

        return this;
    },


    // called by DictProtocol.Listener when it has received data from DictClient
    Write : function (text)
    {
        var stringInputStreamClass = Components.classes[STRINGINPUT_CONTRACTID];

        var inputStream = stringInputStreamClass.getService(nsIInputStream);
        var stringInputStream = inputStream.QueryInterface(nsIStringInputStream);

        stringInputStream.setData(text, text.length);

        this._listener.onStartRequest(this, this._listenerContext);
        this._listener.onDataAvailable(this, this._listenerContext, inputStream, 0, text.length);
        this._listener.onStopRequest(this, this._listenerContext, Components.results.NS_OK);
    },

    //
    // nsIChannel
    //
    loadAttributes        : null,
    contentType           : "text/html",
    contentLength         : 0,
    owner                 : null,
    loadGroup             : null,
    notificationCallbacks : null,
    securityInfo          : null,

    open : function () 
    {
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    },


    asyncOpen : function (observer, context)
    {
        this._listener        = observer;
        this._listenerContext = context;

        var channel = this;

        //   dict://<user>;<auth>@<host>:<port>/d:<word>:<database>:<n>
        //   dict://<user>;<auth>@<host>:<port>/m:<word>:<database>:<strat>:<n>

        var host = channel.URI.host;

        var port = channel.URI.port;
        if (port == -1) {
            // if no port is specified, use the default port
            port = 2628; 
        }

        var path = channel.URI.path;

        const defineUriPattern = /^\/d:([^:]+):?([^:]+)?:?([^:]+)?$/;
        const matchUriPattern  = /^\/m:([^:]+):?([^:]+)?:?([^:]+)?:?([^:]+)?$/;

        var result;
        var word;
        var database;
        var index;
        var strategy;
        var listener;
        var client;

        result = path.match(defineUriPattern);
        if (result != null) {
            // this is a "define" request.
            word     = result[1];
            database = result[2];
            index    = result[3];

            // word is URI-encoded, so we must decode it.
            word = decodeURI(word);

            // According to RFC 2229, Section 5:
            // If database is omitted, "!" SHOULD be used.
            // If index is omitted, then all definitions should be shown.

            if (database == null) {
                database = '!';
            }
            else {
                database = decodeURI(database);
            }

            if (index != null) {
                index = decodeURI(index);
            }

            listener = new DictProtocol.Listener(this, index, host, port, database);
            client   = new DictClient();
            client.defineWord(host, port, word, listener, database);
            return;
        }

        result = path.match(matchUriPattern);
        if (result != null) {
            // this is a "match" request.

            word     = result[1];
            database = result[2];
            strategy = result[3];
            index    = result[4];

            // word is URI-encoded, so we must decode it.
            word = decodeURI(word);

            // According to RFC 2229, Section 5:
            // If database is omitted, "!" SHOULD be used.
            // If strategy is omitted, "." SHOULD be used.
            // If index is omitted, then all definitions should be shown.

            if (database == null) {
                database = '!';
            }
            else {
                database = decodeURI(database);
            }

            if (strategy == null) {
                strategy = '.';
            }
            else {
                strategy = decodeURI(strategy);
            }

            if (index != null) {
                index = decodeURI(index);
            }

            listener = new DictProtocol.Listener(this, index, host, port, database);
            client   = new DictClient();
            client.matchWord(host, port, word, listener, database, strategy);
            return;
        }

        // If we get here, it means that the URI was malformed.
        // TODO: how should we indicate error?
    },

    asyncRead : function (listener, ctxt)
    {
        return listener.onStartRequest (this, ctxt);
    },


    //
    // nsIRequest
    //
    isPending : function ()
    {
        return true;
    },

    cancel : function (status)
    {
        this.status = status;
    },

    suspend : function () 
    {
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    },

    resume : function () 
    {
        throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    }
};


// dict protocol handler 
DictProtocol.Handler = function() {};

DictProtocol.Handler.prototype =
{
    QueryInterface: function(iid)
    {
        if (!iid.equals(nsIProtocolHandler) &&
            !iid.equals(nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
        return this;
    },

    scheme: "dict",

    defaultPort: 2628,

    protocolFlags: nsIProtocolHandler.URI_NORELATIVE |
                   nsIProtocolHandler.URI_NOAUTH     |
                   nsIProtocolHandler.URI_LOADABLE_BY_ANYONE,

    allowPort: function(port, scheme)
    {
        return false;
    },

    newURI: function(spec, charset, baseURI)
    {
        var uri = Components.classes[STANDARDURL_CONTRACTID].createInstance(nsIURI);
        uri.spec = spec;

        return uri;
    },

    newChannel: function(input_uri)
    {
        ios = Components.classes[IOSERVICE_CONTRACTID].getService(nsIIOService);
        if (!ios.allowPort(input_uri.port, input_uri.scheme)) {
            throw Components.results.NS_ERROR_FAILURE;
        }

        return new DictProtocol.Channel(input_uri);
    }
};

// singleton class factory object for DictProtocol objects
DictProtocol.Factory = function() {};

DictProtocol.Factory.prototype = {
    createInstance : function (outer, iid)
    {
        if (outer != null) {
            throw Components.results.NS_ERROR_NO_AGGREGATION;
        }

        if (!iid.equals(nsIProtocolHandler) && 
            !iid.equals(nsISupports)) {
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }

        return new DictProtocol.Handler();
    }
};

// Entry point (Firefox 4 and above)
function NSGetFactory (cid) {
    if (cid.equals(DictProtocol.CLASS_ID)) {
        return new DictProtocol.Factory();
    }
}
